package cz.muni.fi.pb162.cviceni;
import java.util.*;

/**
 * Implementace rozhrani Graph. Neorientovany obycejny graf.
 * 
 * @author Zuzana Neverilova
 * @version 1
 **/
public class GraphImpl implements Graph
{
    private Map<String, Node> nodes;
    private Map<Node, Set<Edge>> nodesedges;

    /**
     * Konstruktor inicializuje kontejner s uzly a s hranami pro kazdy uzel
     */
    public GraphImpl()
    {
        nodes = new HashMap<String, Node>();
        nodesedges = new HashMap<Node, Set<Edge>>();
    }

    /**
     * Prida uzel daneho jmena
     * @param name jmeno uzlu
     * @return uzel, ktery se pridal
     */
    public Node addNode(String name)
    {
        return addNode(new NodeImpl(name));
    }
    
    /**
     * Prida uzel
     * @param node uzel
     * @throws IllegalArgumentException, pokud je uzel uz v grafu
     * @return uzel, ktery se pridal
     */
    public Node addNode(Node node)
    {
        if (nodesedges.containsKey(node))
            throw new IllegalArgumentException("uzel " + node + " uz v grafu je");
        nodesedges.put(node, new HashSet<Edge>());
        nodes.put(node.getName(), node);
        return node;
    }
    
    /**
     * Ziska uzel podle jmena
     * @param name jmeno uzlu
     * @return uzel nebo null, pokud uzel takoveho jmena neexistuje
     */
    public Node getNode(String name){
        if (name == null)
            throw new NullPointerException("name");
        return nodes.get(name);
    }
    
    /**
     * Prida hranu
     * @param edge hrana, ktera se prida
     * @throws NullPointerException, pokud hrana nema uzly v grafu
     * @return hrana, ktera se pridala
     */
    public Edge addEdge(Edge edge)
    {
        if (!nodesedges.containsKey(edge.getFrom()))
            throw new NullPointerException("edge.getFrom()");
        if (!nodesedges.containsKey(edge.getTo()))
            throw new NullPointerException("edge.getTo()");
        nodesedges.get(edge.getFrom()).add(edge);
        nodesedges.get(edge.getTo()).add(edge);  // neorientovany graf      
        return edge;
    }
    
    /**
     * Vytvori a prida hranu mezi 2 uzly o zadane delce. Pokud nektery z uzlu v grafu neni, prida se.
     * @param from jmeno uzlu, ze ktereho hrana vychazi
     * @param to jmeno uzlu, kam hrana smeruje
     * @param length delka hrany
     * @return pridana hrana
     */
    public Edge addEdge(String from, String to, int length)
    {
        if (!nodesedges.containsKey(getNode(from)))
            addNode(from);
        if (!nodesedges.containsKey(getNode(to)))
            addNode(to);            
        return addEdge(new EdgeImpl(getNode(from), getNode(to), length));
    }
    
    /**
     * Vrati mnozinu hran v grafu, ktere vedou ze zadaneho uzlu.
     * @param from uzel, ze ktereho hrany vychazeji
     * @return mnozina hran
     */
    public Set<Edge> getEdges(Node from)
    {
        return nodesedges.get(from);
    }
    
    /**
     * Prida vsechny uzly z kontejneru.
     * @param nodes kontejner<String, Node> s uzly
     */
    public void addNodes(Map<String, Node> nodes){
        this.nodes.putAll(nodes);
        for (Iterator <Node> i = nodes.values().iterator(); i.hasNext();)
        {
            Node n = i.next();
            if (!nodesedges.containsKey(n))
                nodesedges.put(n, new HashSet<Edge>());
        }
    }
    
    /**
     * Vrati true, pokud existuje cesta z jednoho uzlu do druheho.
     * @param from uzel, ze ktereho vychazime
     * @param to uzel, do ktereho vchazime
     * @return true, pokud cesta existuje, jinak false
     * @throws NullPointerException, pokud je nektery zadany uzel null
     * @throws IllegalArgumentException, pokud nektery uzel neexistuje
     */
    public boolean existsPath(String from, String to)
    {
        if (from == null)
            throw new NullPointerException("from");
        if (to == null)
            throw new NullPointerException("to");
        if (getNode(from) == null)
            throw new IllegalArgumentException("uzel " + from + " v grafu neni ");
        if (getNode(to) == null)
            throw new IllegalArgumentException("uzel " + to + " v grafu neni ");
        Set <Node> visited = new HashSet<Node>();
        visited.add(getNode(from));
        return findPath(getNode(to), visited);
    }
    
    /**
     * Rekurzivni metoda pro hledani cesty v grafu. Pamatuji se jiz navstivene uzly (v promenne
     * visited) a prochazi se jejich seznam a vsechny hrany, ktere odtud vedou. Pokud se narazi
     * na hledany uzel, vrati se true a rekurze nepokracuje. Pokud se projdou oba iteratory,
     * znamena to, ze cesta nebyla nalezena.
     * @param to hledany uzel
     * @param visited jiz navstivene uzly
     * @return true, pokud existuje cesta z visited do to
     */
    private boolean findPath(Node to, Set<Node> visited){
         Set <Node> newvisited = new HashSet<Node>();
         newvisited.addAll(visited);
         for (Iterator <Node> j = visited.iterator(); j.hasNext(); ){
            Set<Edge> edges = getEdges(j.next());
            for (Iterator <Edge> i = edges.iterator(); i.hasNext(); )
            {
                Edge e = i.next();
                if (e.getTo().equals(to)||e.getFrom().equals(to))
                    return true;
                if (!visited.contains(e.getTo())){
                    newvisited.add(e.getTo());
                    if (findPath(to, newvisited))
                        return true;
                }
                if (!visited.contains(e.getFrom())){ // neorientovany graf
                    newvisited.add(e.getFrom());
                    if (findPath(to, newvisited))
                        return true;
                }            
            }
        }
        return false;
    }
    
    /**
     * Textova reprezentace grafu. Vypisou se hrany (mohly by se vypsat i uzly).
     * @return textova reprezentace grafu
     */
    public String toString()
    {
        return getEdges().toString();
    }
    
    /**
     * Vrati vsechny hrany grafu.
     * @return mnozina vsech hran
     */
    public Set<Edge> getEdges()
    {
        Set <Edge> edges = new HashSet<Edge>();
        for (Iterator <Set<Edge>> i = nodesedges.values().iterator(); i.hasNext(); )
        {
            edges.addAll(i.next());
        }
        return edges;
    }
    
    /**
     * Vrati graf, ktery je minimalni kostrou grafu. Alogoritmus nalezeni minimalni kostry
     * je Kruskaluv-Boruvkuv.
     * @return minimalni kostra
     */
    public Graph minimumSpanTree()
    {
        Graph minimumSpanTree = new GraphImpl();
        minimumSpanTree.addNodes(nodes);
        SortedSet<Edge> edges = new TreeSet<Edge>(getEdges());
        for (Iterator<Edge> i = edges.iterator(); i.hasNext(); )
        {
            Edge e = i.next();
            if (!minimumSpanTree.existsPath(e.getFrom().getName(), e.getTo().getName()))
                minimumSpanTree.addEdge(e);
        }
        return minimumSpanTree;
    }
}
